// This file is part of "Notepad++ C++ Plugins" template.
// Copyright 2025 by Randall Joseph Fellmy <software@coises.com>, <http://www.coises.com/software/>

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


// ** Some inline utility / helper routines -- these are for convenience; none are required by Notepad++ or Scintilla **

#pragma once
#include "PluginFramework.h"


// Sending messages to Notepad++ and to Scintilla

template<typename W, typename L> LRESULT npp(UINT Msg, W wParam, L lParam) {
    if constexpr (std::is_integral_v<W>)
        if constexpr (std::is_integral_v<L>)
             return SendMessage(plugin.nppData._nppHandle, Msg, static_cast<WPARAM>(wParam), static_cast<LPARAM>(lParam));
        else return SendMessage(plugin.nppData._nppHandle, Msg, static_cast<WPARAM>(wParam), reinterpret_cast<LPARAM>(lParam));
    else if constexpr (std::is_integral_v<L>)
         return SendMessage(plugin.nppData._nppHandle, Msg, reinterpret_cast<WPARAM>(wParam), static_cast<LPARAM>(lParam));
    else return SendMessage(plugin.nppData._nppHandle, Msg, reinterpret_cast<WPARAM>(wParam), reinterpret_cast<LPARAM>(lParam));
}

inline Scintilla::ScintillaCall& sci = plugin.sci;


// Convert between Windows utf-16 strings and Scintilla's ANSI or utf-8 strings, even if they are very long.

inline std::string fromWide(std::wstring_view s, unsigned int codepage) {
    std::string r;
    size_t inputLength = s.length();
    if (!inputLength) return r;
    constexpr unsigned int safeSize = std::numeric_limits<int>::max() / 8;
    size_t workingPoint = 0;
    while (inputLength - workingPoint > safeSize) {
        int ss = safeSize;
        if (s[ss - 1] >= 0xD800 && s[ss - 1] <= 0xDBFF) --ss;  // proposed block ends with high surrogate: leave it for the next block
        int segmentLength = WideCharToMultiByte(codepage, 0, s.data() + workingPoint, ss, 0, 0, 0, 0);
        size_t outputPoint = r.length();
        r.resize(outputPoint + segmentLength);
        WideCharToMultiByte(codepage, 0, s.data() + workingPoint, ss, r.data() + outputPoint, segmentLength, 0, 0);
        workingPoint += ss;
    }
    int segmentLength = WideCharToMultiByte(codepage, 0, s.data() + workingPoint, static_cast<int>(inputLength - workingPoint), 0, 0, 0, 0);
    size_t outputPoint = r.length();
    r.resize(outputPoint + segmentLength);
    WideCharToMultiByte(codepage, 0, s.data() + workingPoint, static_cast<int>(inputLength - workingPoint), r.data() + outputPoint, segmentLength, 0, 0);
    return r;
}

inline std::wstring toWide(std::string_view s, unsigned int codepage) {
    std::wstring r;
    size_t inputLength = s.length();
    if (!inputLength) return r;
    constexpr unsigned int safeSize = std::numeric_limits<int>::max() / 2;
    size_t workingPoint = 0;
    while (inputLength - workingPoint > safeSize) {
        int ss = safeSize;
        if (codepage == CP_UTF8 && ((s[ss] & 0xC0) == 0x80)) while ((s[--ss] & 0xC0) == 0x80);  // find a first byte to start the next block
        int segmentLength = MultiByteToWideChar(codepage, 0, s.data() + workingPoint, ss, 0, 0);
        size_t outputPoint = r.length();
        r.resize(outputPoint + segmentLength);
        MultiByteToWideChar(codepage, 0, s.data() + workingPoint, ss, r.data() + outputPoint, segmentLength);
        workingPoint += ss;
    }
    int segmentLength = MultiByteToWideChar(codepage, 0, s.data() + workingPoint, static_cast<int>(inputLength - workingPoint), 0, 0);
    size_t outputPoint = r.length();
    r.resize(outputPoint + segmentLength);
    MultiByteToWideChar(codepage, 0, s.data() + workingPoint, static_cast<int>(inputLength - workingPoint), r.data() + outputPoint, segmentLength);
    return r;
}

inline std::string  fromWide(std::wstring_view s) { return fromWide(s, plugin.sci.CodePage()); }
inline std::wstring toWide  (std::string_view  s) { return   toWide(s, plugin.sci.CodePage()); }


// Get the text from a Window or a from a control in a dialog as a std::wstring

inline std::wstring GetWindowString(HWND hWnd) {
    std::wstring s(GetWindowTextLength(hWnd), 0);
    if (!s.empty()) s.resize(GetWindowText(hWnd, s.data(), static_cast<int>(s.length() + 1)));
    return s;
}

inline std::wstring GetDlgItemString(HWND hwndDlg, int item) {
    return GetWindowString(GetDlgItem(hwndDlg, item));
}


// Show a balloon tip on an edit control, or on the edit control associated with a combobox or a spin control, in a dialog.

inline bool ShowBalloonTip(HWND hwndDlg, int item, const std::wstring& text) {
    HWND hControl = GetDlgItem(hwndDlg, item);
    std::wstring classname(127, 0);
    GetClassName(hControl, classname.data(), static_cast<int>(classname.length() + 1));
    classname.resize(std::min(classname.find(L'\0'), classname.length()));
    if (classname == UPDOWN_CLASS) {
        hControl = reinterpret_cast<HWND>(SendMessage(hControl, UDM_GETBUDDY, 0, 0));
        if (!hControl) return false;
    }
    else if (classname == WC_COMBOBOX || classname == WC_COMBOBOXEX) {
        COMBOBOXINFO cbi;
        cbi.cbSize = sizeof(COMBOBOXINFO);
        if (!GetComboBoxInfo(hControl, &cbi)) return false;
        hControl = cbi.hwndItem;
    }
    EDITBALLOONTIP ebt;
    ebt.cbStruct = sizeof(EDITBALLOONTIP);
    ebt.pszTitle = L"";
    ebt.ttiIcon = TTI_NONE;
    ebt.pszText = text.data();
    if (!SendMessage(hControl, EM_SHOWBALLOONTIP, 0, reinterpret_cast<LPARAM>(&ebt))) return false;
    SendMessage(hwndDlg, WM_NEXTDLGCTL, reinterpret_cast<WPARAM>(hControl), TRUE);
    return true;
}


// Get the full path of a file open in Notepad++ as a std::wstring.
// If an argument is supplied, it is the Notepad++ buffer id to be examined.
// If no argument is given, the current buffer is examined; this only works in commands and NPPN_BUFFERACTIVATED notifications,
// since in Scintilla notifications and most Notepad++ notifications, the "current buffer" is not necessarily meaningful.

inline std::wstring getFilePath(UINT_PTR buffer) {
    auto n = SendMessage(plugin.nppData._nppHandle, NPPM_GETFULLPATHFROMBUFFERID, buffer, 0);
    if (n < 1) return L"";
    std::wstring s(n, 0);
    SendMessage(plugin.nppData._nppHandle, NPPM_GETFULLPATHFROMBUFFERID, buffer, (LPARAM)s.data());
    return s;
}
inline std::wstring getFilePath() { return getFilePath(SendMessage(plugin.nppData._nppHandle, NPPM_GETCURRENTBUFFERID, 0, 0)); }


// Get the lower case file extension from a path, a buffer id or the current buffer.
// Returns L"." for a file with no extension, L"" for a non-path (i.e., without a backslash, e.g., "new 1").
// If no argument is given, the current buffer is examined; this only works in commands and NPPN_BUFFERACTIVATED notifications,
// since in Scintilla notifications and most Notepad++ notifications, the "current buffer" is not necessarily meaningful.

inline std::wstring getFileExtension(const std::wstring& filepath) {
    size_t lastBackslash = filepath.find_last_of(L'\\');
    if (lastBackslash == std::wstring::npos) return L"";
    size_t lastPeriod = filepath.find_last_of(L'.');
    if (lastPeriod == std::wstring::npos || lastPeriod < lastBackslash || lastPeriod == filepath.length() - 1) return L".";
    std::wstring ext = filepath.substr(lastPeriod + 1);
    wcslwr(ext.data());
    return ext;
}
inline std::wstring getFileExtension(UINT_PTR buffer) { return getFileExtension(getFilePath(buffer)); }
inline std::wstring getFileExtension() { return getFileExtension(SendMessage(plugin.nppData._nppHandle, NPPM_GETCURRENTBUFFERID, 0, 0)); }


// A struct for managing resizable dialogs

struct DialogStretch {

    HWND dialog = 0;
    RECT original = { 0, 0, 0, 0 };
    std::map<HWND, RECT> controls;

    DialogStretch() {}

    void setup(HWND hwndDlg) {
        dialog = hwndDlg;
        GetWindowRect(dialog, &original);
        EnumChildWindows(dialog, [](HWND h, LPARAM p) -> BOOL {
            (*reinterpret_cast<std::map<HWND, RECT>*>(p)).emplace(h, RECT());
            return TRUE;
        }, reinterpret_cast<LPARAM>(&controls));
        for (auto& c : controls) {
            GetWindowRect(c.first, &c.second);
            MapWindowPoints(0, dialog, reinterpret_cast<LPPOINT>(&c.second), 2);
        }
    }

    int originalWidth () const { return original.right - original.left; }
    int originalHeight() const { return original.bottom - original.top; }

    class Stretched {
        friend struct DialogStretch;
        HWND dialog;
        const std::map<HWND, RECT>& controls;
        int addWidth = 0;
        int addHeight = 0;
        Stretched(HWND dialog, const RECT& original, const std::map<HWND, RECT>& controls) : dialog(dialog), controls(controls) {
            RECT currentRectangle;
            GetWindowRect(dialog, &currentRectangle);
            addWidth = currentRectangle.right - currentRectangle.left - original.right + original.left;
            addHeight = currentRectangle.bottom - currentRectangle.top - original.bottom + original.top;
        }
    public:
        Stretched& adjust(HWND h, double xStretch, double yStretch = 0, double xMove = 0, double yMove = 0) {
            if (controls.contains(h)) {
                const RECT& cRef = controls.at(h);
                SetWindowPos(h, 0, std::lround(cRef.left + xMove * addWidth),
                    std::lround(cRef.top + yMove * addHeight),
                    std::lround(cRef.right - cRef.left + xStretch * addWidth),
                    std::lround(cRef.bottom - cRef.top + yStretch * addHeight),
                    SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_NOZORDER);
                InvalidateRect(h, 0, TRUE);
            }
            return *this;
        }
        Stretched& adjust(int control, double xStretch, double yStretch = 0, double xMove = 0, double yMove = 0) {
            return adjust(GetDlgItem(dialog, control), xStretch, yStretch, xMove, yMove);
        }
    };

    Stretched adjust(HWND h, double xStretch, double yStretch = 0, double xMove = 0, double yMove = 0) {
        Stretched stretched(dialog, original, controls);
        return stretched.adjust(h, xStretch, yStretch, xMove, yMove);
    }

    Stretched adjust(int control, double xStretch, double yStretch = 0, double xMove = 0, double yMove = 0) {
        Stretched stretched(dialog, original, controls);
        return stretched.adjust(control, xStretch, yStretch, xMove, yMove);
    }

};